@@ -20,7 +20,11 @@ from azure.mgmt.sql.models import (
     CreateMode,
     DatabaseEdition,
     EncryptionProtector,
+    FailoverGroup,
+    FailoverGroupReadOnlyEndpoint,
+    FailoverGroupReadWriteEndpoint,
     IdentityType,
+    PartnerInfo,
     PerformanceLevelUnit,
     ReplicationRole,
     ResourceIdentity,
@@ -81,6 +85,29 @@ def _any_sku_values_specified(sku):
     return any(val for key, val in sku.__dict__.items())
 
 
+def _get_default_server_version(location_capabilities):
+    '''
+    Gets the default server version capability from the full location
+    capabilities response.
+
+    If none have 'default' status, gets the first capability that has
+    'available' status.
+
+    If there is no default or available server version, falls back to
+    server version 12.0 in order to maintain compatibility with older
+    Azure CLI releases.
+    '''
+    server_versions = location_capabilities.supported_server_versions
+
+    try:
+        # Try behavior from azure-cli-sql 2.0.26: get default
+        return _get_default_capability(server_versions)
+    except StopIteration:
+        # No default or available version found.
+        # Fall back to behaviour from azure-cli-sql 2.0.25 and earlier: get version 12.0
+        return next(sv for sv in server_versions if sv.name == "12.0")
+
+
 def _get_default_capability(capabilities):
     '''
     Gets the first capability in the collection that has 'default' status.
@@ -96,7 +123,7 @@ def is_available(status):
     Returns True if the capability status is available (including default).
     '''
 
-    return status != CapabilityStatus.visible and status != CapabilityStatus.visible.value
+    return status not in (CapabilityStatus.visible, CapabilityStatus.visible.value)
 
 
 def _filter_available(capabilities):
@@ -413,7 +440,7 @@ def db_show_conn_str(
     return f.format(**conn_str_props)
 
 
-class DatabaseIdentity(object):  # pylint: disable=too-few-public-methods
+class DatabaseIdentity():  # pylint: disable=too-few-public-methods
     '''
     Helper class to bundle up database identity properties and generate
     database resource id.
@@ -464,7 +491,7 @@ def _find_db_sku_from_capabilities(cli_ctx, location, sku, allow_reset_family=Fa
     # Get default server version capability
     capabilities_client = get_sql_capabilities_operations(cli_ctx, None)
     capabilities = capabilities_client.list_by_location(location, CapabilityGroup.supported_editions)
-    server_version_capability = _get_default_capability(capabilities.supported_server_versions)
+    server_version_capability = _get_default_server_version(capabilities)
 
     # Find edition capability, based on requested sku properties
     edition_capability = _find_edition_capability(
@@ -483,6 +510,38 @@ def _find_db_sku_from_capabilities(cli_ctx, location, sku, allow_reset_family=Fa
     return result
 
 
+def _validate_elastic_pool_id(
+        cli_ctx,
+        elastic_pool_id,
+        server_name,
+        resource_group_name):
+    '''
+    Validates elastic_pool_id is either None or a valid resource id.
+
+    If elastic_pool_id has a value but it is not a valid resource id,
+    then assume that user specified elastic pool name which we need to
+    convert to elastic pool id using the provided server & resource group
+    name.
+
+    Returns the elastic_pool_id, which may have been updated and may be None.
+    '''
+
+    from msrestazure.tools import resource_id, is_valid_resource_id
+    from azure.cli.core.commands.client_factory import get_subscription_id
+
+    if elastic_pool_id and not is_valid_resource_id(elastic_pool_id):
+        return resource_id(
+            subscription=get_subscription_id(cli_ctx),
+            resource_group=resource_group_name,
+            namespace='Microsoft.Sql',
+            type='servers',
+            name=server_name,
+            child_type_1='elasticPools',
+            child_name_1=elastic_pool_id)
+
+    return elastic_pool_id
+
+
 def _db_dw_create(
         cli_ctx,
         client,
@@ -510,6 +569,13 @@ def _db_dw_create(
     # using capabilities.
     kwargs['sku'] = _find_db_sku_from_capabilities(cli_ctx, kwargs['location'], sku)
 
+    # Validate elastic pool id
+    kwargs['elastic_pool_id'] = _validate_elastic_pool_id(
+        cli_ctx,
+        kwargs['elastic_pool_id'],
+        dest_db.server_name,
+        dest_db.resource_group_name)
+
     # Create
     return sdk_no_wait(no_wait, client.create_or_update,
                        server_name=dest_db.server_name,
@@ -778,7 +844,7 @@ def db_list_capabilities(
     capabilities = client.list_by_location(location, CapabilityGroup.supported_editions)
 
     # Get subtree related to databases
-    editions = _get_default_capability(capabilities.supported_server_versions).supported_editions
+    editions = _get_default_server_version(capabilities).supported_editions
 
     # Filter by edition
     if edition:
@@ -1383,7 +1449,7 @@ def _find_elastic_pool_sku_from_capabilities(cli_ctx, location, sku, allow_reset
     # Get default server version capability
     capabilities_client = get_sql_capabilities_operations(cli_ctx, None)
     capabilities = capabilities_client.list_by_location(location, CapabilityGroup.supported_elastic_pool_editions)
-    server_version_capability = _get_default_capability(capabilities.supported_server_versions)
+    server_version_capability = _get_default_server_version(capabilities)
 
     # Find edition capability, based on requested sku properties
     edition_capability = _find_edition_capability(sku, server_version_capability.supported_elastic_pool_editions)
@@ -1509,7 +1575,7 @@ def elastic_pool_list_capabilities(
     capabilities = client.list_by_location(location, CapabilityGroup.supported_elastic_pool_editions)
 
     # Get subtree related to elastic pools
-    editions = _get_default_capability(capabilities.supported_server_versions).supported_elastic_pool_editions
+    editions = _get_default_server_version(capabilities).supported_elastic_pool_editions
 
     # Filter by edition
     if edition:
@@ -2048,3 +2114,167 @@ def managed_db_restore(
         managed_instance_name=target_managed_instance_name,
         resource_group_name=target_resource_group_name,
         parameters=kwargs)
+
+###############################################
+#              sql failover-group             #
+###############################################
+
+
+# pylint: disable=too-few-public-methods
+class FailoverPolicyType(Enum):
+    automatic = 'Automatic'
+    manual = 'Manual'
+
+
+def failover_group_create(
+        cmd,
+        client,
+        resource_group_name,
+        server_name,
+        failover_group_name,
+        partner_server,
+        partner_resource_group=None,
+        failover_policy=FailoverPolicyType.automatic.value,
+        grace_period=1,
+        add_db=None):
+    '''
+    Creates a failover group.
+    '''
+
+    from six.moves.urllib.parse import quote  # pylint: disable=import-error
+    from azure.cli.core.commands.client_factory import get_subscription_id
+
+    # Build the partner server id
+    partner_server_id = "/subscriptions/{}/resourceGroups/{}/providers/Microsoft.Sql/servers/{}".format(
+        quote(get_subscription_id(cmd.cli_ctx)),
+        quote(partner_resource_group or resource_group_name),
+        quote(partner_server))
+
+    partner_server = PartnerInfo(id=partner_server_id)
+
+    # Convert grace period from hours to minutes
+    grace_period = int(grace_period) * 60
+
+    if failover_policy == FailoverPolicyType.manual.value:
+        grace_period = None
+
+    if add_db is None:
+        add_db = []
+
+    databases = _get_list_of_databases_for_fg(
+        cmd,
+        resource_group_name,
+        server_name,
+        [],
+        add_db,
+        [])
+
+    return client.create_or_update(
+        resource_group_name=resource_group_name,
+        server_name=server_name,
+        failover_group_name=failover_group_name,
+        parameters=FailoverGroup(
+            partner_servers=[partner_server],
+            databases=databases,
+            read_write_endpoint=FailoverGroupReadWriteEndpoint(
+                failover_policy=failover_policy,
+                failover_with_data_loss_grace_period_minutes=grace_period),
+            read_only_endpoint=FailoverGroupReadOnlyEndpoint(
+                failover_policy="Disabled")))
+
+
+def failover_group_update(
+        cmd,
+        instance,
+        resource_group_name,
+        server_name,
+        failover_policy=None,
+        grace_period=None,
+        add_db=None,
+        remove_db=None):
+    '''
+    Updates the failover group.
+    '''
+
+    if failover_policy is not None:
+        instance.read_write_endpoint.failover_policy = failover_policy
+
+    if instance.read_write_endpoint.failover_policy == FailoverPolicyType.manual.value:
+        grace_period = None
+        instance.read_write_endpoint.failover_with_data_loss_grace_period_minutes = grace_period
+
+    if grace_period is not None:
+        grace_period = int(grace_period) * 60
+        instance.read_write_endpoint.failover_with_data_loss_grace_period_minutes = grace_period
+
+    if add_db is None:
+        add_db = []
+
+    if remove_db is None:
+        remove_db = []
+
+    databases = _get_list_of_databases_for_fg(
+        cmd,
+        resource_group_name,
+        server_name,
+        instance.databases,
+        add_db,
+        remove_db)
+
+    instance.databases = databases
+
+    return instance
+
+
+def failover_group_failover(
+        client,
+        resource_group_name,
+        server_name,
+        failover_group_name,
+        allow_data_loss=False):
+    '''
+    Failover a failover group.
+    '''
+
+    failover_group = client.get(
+        resource_group_name=resource_group_name,
+        server_name=server_name,
+        failover_group_name=failover_group_name)
+
+    if failover_group.replication_role == "Primary":
+        return
+
+    # Choose which failover method to use
+    if allow_data_loss:
+        failover_func = client.force_failover_allow_data_loss
+    else:
+        failover_func = client.failover
+
+    return failover_func(
+        resource_group_name=resource_group_name,
+        server_name=server_name,
+        failover_group_name=failover_group_name)
+
+
+def _get_list_of_databases_for_fg(
+        cmd,
+        resource_group_name,
+        server_name,
+        databases_in_fg,
+        add_db,
+        remove_db):
+    '''
+    Gets a list of databases that are supposed to be part of the failover group
+    after the operation finishes
+    It consolidates the list of dbs to add and remove with the list of databases
+    that are already part of the failover group.
+    '''
+
+    add_db_ids = [DatabaseIdentity(cmd.cli_ctx, d, server_name, resource_group_name).id() for d in add_db]
+
+    remove_db_ids = [DatabaseIdentity(cmd.cli_ctx, d, server_name, resource_group_name).id() for d in remove_db]
+
+    databases = list(({x.lower() for x in databases_in_fg} |
+                      {x.lower() for x in add_db_ids}) - {x.lower() for x in remove_db_ids})
+
+    return databases
